@@ -37,6 +37,7 @@
 #include <sys/un.h>
 
 #include <stdlib.h>
+#include <ctype.h>
 #include <stdint.h>
 #include <pthread.h>
 #include <unistd.h>
@@ -172,6 +173,100 @@ gkr_memory_fallback (void *p, unsigned long sz)
 	return realloc (p, sz);	
 }
 
+/* -----------------------------------------------------------------------------
+ * MODULE ARGUMENTS
+ */
+
+static void
+parse_argument (char *arg)
+{
+	char *value;
+	
+	value = arg + strcspn (arg, ":=");
+	if (!*value)
+		value = NULL;
+	else 
+		*(value++) = 0;
+
+	/* Setup the socket path from the arguments */
+	if (strcmp (arg, "socket") == 0) {
+		strncpy (socket_path, value, sizeof (socket_path));
+		socket_path[sizeof (socket_path) - 1] = 0;		
+	} else {
+		WARN(("unrecognized argument: %s", arg));
+	}
+}
+
+static void
+parse_arguments (const char *string) 
+{
+	char quote = '\0';
+	char *src, *dup, *at, *arg;
+	
+	if (!string)
+		return;
+	
+	src = dup = strdup (string);
+	if (!dup) {
+		WARN (("couldn't allocate memory for argument string"));
+		return;
+	}
+
+	arg = at = src;
+	for (src = dup; *src; src++) {
+		
+		/* Matching quote */
+		if (quote == *src) {
+			quote = '\0';
+			
+		/* Inside of quotes */
+		} else if (quote != '\0') {
+			if (*src == '\\') {
+				*at++ = *src++;
+				if (!*src) {
+					WARN (("couldn't parse argument string: %s", string));
+					goto done;
+				}
+				if (*src != quote) 
+					*at++ = '\\';
+			}
+			*at++ = *src;
+			
+		/* Space, not inside of quotes */
+		} else if (isspace(*src)) {
+			*at = 0;
+			parse_argument (arg);
+			arg = at;
+			
+		/* Other character outside of quotes */
+		} else {
+			switch (*src) {
+			case '\'':
+			case '"':
+				quote = *src;
+				break;
+			case '\\':
+				*at++ = *src++;
+				if (!*src) {
+					WARN (("couldn't parse argument string: %s", string));
+					goto done;
+				}
+				/* fall through */
+			default:
+				*at++ = *src;
+				break;
+			}
+		}
+	}
+
+	
+	if (at != arg) 
+		parse_argument (arg);
+	
+done:
+	free (dup);
+}
+
 /* -----------------------------------------------------------------------------
  * CALL SESSION
  */
@@ -978,7 +1073,7 @@ call_session_close_all ()
 
 static CK_RV
 proto_read_attribute_array (GkrPkcs11Message *msg, CK_ATTRIBUTE_PTR arr, 
-                            CK_ULONG_PTR len, CK_ULONG max)
+                            CK_ULONG_PTR len, CK_ULONG max, int *overflowed)
 {
 	uint32_t i, num, val;
 	CK_ATTRIBUTE_PTR attr;
@@ -1034,10 +1129,12 @@ proto_read_attribute_array (GkrPkcs11Message *msg, CK_ATTRIBUTE_PTR arr,
 				/* Just requesting the attribute size */
 				if (!attr->pValue) {
 					attr->ulValueLen = attrlen;
+					(*overflowed)++;
 
 				/* Wants attribute data, but too small */
 				} else if (attr->ulValueLen < attrlen) {
 					attr->ulValueLen = attrlen;
+					(*overflowed)++;
 					ret = CKR_BUFFER_TOO_SMALL;
 
 				/* Wants attribute data, value is null */
@@ -1062,7 +1159,7 @@ proto_read_attribute_array (GkrPkcs11Message *msg, CK_ATTRIBUTE_PTR arr,
 
 static CK_RV
 proto_read_byte_array (GkrPkcs11Message *msg, CK_BYTE_PTR arr,
-                       CK_ULONG_PTR len, CK_ULONG max)
+                       CK_ULONG_PTR len, CK_ULONG max, int *overflowed)
 {
 	const unsigned char *val;
 	size_t vlen;
@@ -1081,11 +1178,15 @@ proto_read_byte_array (GkrPkcs11Message *msg, CK_BYTE_PTR arr,
 	*len = vlen;
 
 	/* Just asking us for size */
-	if (!arr) 
+	if (!arr) {
+		(*overflowed)++;
 		return CKR_OK;
+	}
 
-	if (max < vlen)
+	if (max < vlen) {
+		(*overflowed)++;
 		return CKR_BUFFER_TOO_SMALL;
+	}
 
 	/* Enough space, yay */
 	memcpy (arr, val, vlen);
@@ -1095,7 +1196,7 @@ proto_read_byte_array (GkrPkcs11Message *msg, CK_BYTE_PTR arr,
 
 static CK_RV
 proto_read_uint32_array (GkrPkcs11Message *msg, CK_ULONG_PTR arr,
-                         CK_ULONG_PTR len, CK_ULONG max)
+                         CK_ULONG_PTR len, CK_ULONG max, int *overflowed)
 {
 	uint32_t i, num, val;
 
@@ -1112,8 +1213,12 @@ proto_read_uint32_array (GkrPkcs11Message *msg, CK_ULONG_PTR arr,
 
 	*len = num;
 
-	if (arr && max < num)
+	if (!arr) {
+		(*overflowed)++;
+	} else if (max < num) {
+		(*overflowed)++;
 		return CKR_BUFFER_TOO_SMALL;
+	}
 
 	/* We need to go ahead and read everything in all cases */
 	for (i = 0; i < num; ++i) {
@@ -1179,7 +1284,7 @@ proto_read_sesssion_info (GkrPkcs11Message *msg, CK_SESSION_INFO_PTR info)
 	
 	/* The slot id (we ignore) */
 	gkr_buffer_get_uint32 (&msg->buffer, msg->parsed, &msg->parsed, &val);
-	info->slotID = 0;
+	info->slotID = slot_id;
 
 	/* The state */
 	gkr_buffer_get_uint32 (&msg->buffer, msg->parsed, &msg->parsed, &val);
@@ -1208,7 +1313,6 @@ gkr_C_Initialize (CK_VOID_PTR init_args)
 	CK_RV ret = CKR_OK;
 	pid_t initialize_pid;
 	char *path;
-	int l;
 	
 	DBG (("C_Initialize: enter"));
 
@@ -1226,7 +1330,6 @@ gkr_C_Initialize (CK_VOID_PTR init_args)
 
 		/* pReserved must be NULL */
 		args = init_args;
-		PREREQ (!args->pReserved, CKR_ARGUMENTS_BAD);
 
 		/* ALL supplied function pointers need to have the value either NULL or non-NULL. */
 		supplied_ok = (args->CreateMutex == NULL && args->DestroyMutex == NULL &&
@@ -1237,11 +1340,18 @@ gkr_C_Initialize (CK_VOID_PTR init_args)
 
 		/*
 		 * When the CKF_OS_LOCKING_OK flag isn't set and mutex function pointers are supplied
-		 * by an application, return an error.  DBus must be able to use its own locks.
+		 * by an application, return an error.  We must be able to use its own locks.
 		 */
 		if (!(args->flags & CKF_OS_LOCKING_OK) && (args->CreateMutex != NULL)) {
 			PREREQ (FALSE, CKR_CANT_LOCK);
 		}
+		
+		/* 
+		 * We support setting the socket path and other arguments from from the 
+		 * pReserved pointer, similar to how NSS PKCS#11 components are initialized. 
+		 */ 
+		if (args->pReserved) 
+			parse_arguments ((const char*)args->pReserved);			
 	}
 
 	/* Main initialization */
@@ -1267,16 +1377,16 @@ gkr_C_Initialize (CK_VOID_PTR init_args)
 			crypto_pid = initialize_pid;
 			
 			/* Lookup the socket path, append '.pkcs11' */
-			socket_path[0] = 0;
-			path = getenv ("GNOME_KEYRING_SOCKET");
-			if (path && path[0]) {
-				l = sizeof (socket_path) - 1;
-				strncpy (socket_path, path, l);
-				strncat (socket_path, ".pkcs11", l);
-				socket_path[l] = 0;
-				
-				DBG (("gnome-keyring pkcs11 socket is: %s", socket_path));
+			if (socket_path[0] == 0) {
+				socket_path[0] = 0;
+				path = getenv ("GNOME_KEYRING_SOCKET");
+				if (path && path[0]) {
+					snprintf (socket_path, sizeof (socket_path), "%s.pkcs11", path);
+					socket_path[sizeof (socket_path) - 1] = 0;
+				}
 			}
+			
+			DBG (("gnome-keyring pkcs11 socket is: %s", socket_path));
 		}
 
 	pthread_mutex_unlock (&global_mutex);
@@ -1597,7 +1707,7 @@ gkr_C_OpenSession (CK_SLOT_ID id, CK_FLAGS flags, CK_VOID_PTR user_data,
 	DBG (("C_OpenSession: enter"));
 	PREREQ (pkcs11_initialized, CKR_CRYPTOKI_NOT_INITIALIZED);
 	PREREQ (session, CKR_ARGUMENTS_BAD);
-	PREREQ (flags & CKF_SERIAL_SESSION, CKR_FUNCTION_NOT_PARALLEL);
+	PREREQ (flags & CKF_SERIAL_SESSION, CKR_SESSION_PARALLEL_NOT_SUPPORTED);
 
 	if (id != slot_id) 
 		return CKR_SLOT_ID_INVALID;
@@ -1715,7 +1825,7 @@ gkr_C_CloseAllSessions (CK_SLOT_ID id)
 		ret = CKR_SLOT_ID_INVALID;
 	
 	DBG (("C_CloseAllSessions: %d", ret));
-	return CKR_OK;
+	return ret;
 }
 
 static CK_RV
@@ -1764,6 +1874,7 @@ gkr_C_CancelFunction (CK_SESSION_HANDLE hSession)
 	} 
 
 #define IN_ATTRIBUTE_ARRAY(arr, num) \
+	if (num && !arr) { _ret = CKR_ARGUMENTS_BAD; goto _cleanup; } \
 	_ret = gkr_pkcs11_message_write_attribute_array (_cs->req, (arr), (num)); \
 	if (_ret != CKR_OK) goto _cleanup;
 
@@ -1784,13 +1895,13 @@ gkr_C_CancelFunction (CK_SESSION_HANDLE hSession)
 	if (_ret != CKR_OK) goto _cleanup;
 
 #define OUT_ATTRIBUTE_ARRAY(arr, num, max) \
-	if (!arr) _cs->overflowed = 1; \
-	_ret = proto_read_attribute_array (_cs->resp, (arr), (num), (max)); \
+	if (!arr) _cs->overflowed++; \
+	_ret = proto_read_attribute_array (_cs->resp, (arr), (num), (max), &_cs->overflowed); \
 	if (_ret != CKR_OK) goto _cleanup;
 
 #define OUT_BYTE_ARRAY(arr, len, max)  \
-	if (!arr) _cs->overflowed = 1; \
-	_ret = proto_read_byte_array (_cs->resp, (arr), (len), (max)); \
+	if (!arr) _cs->overflowed++; \
+	_ret = proto_read_byte_array (_cs->resp, (arr), (len), (max), &_cs->overflowed); \
 	if (_ret != CKR_OK) goto _cleanup;
 
 #define OUT_HANDLE(val) \
@@ -1798,8 +1909,9 @@ gkr_C_CancelFunction (CK_SESSION_HANDLE hSession)
 	if (_ret != CKR_OK) goto _cleanup;
 
 #define OUT_HANDLE_ARRAY(a, n, mx) \
-	if (!a) _cs->overflowed = 1; \
-	_ret = proto_read_uint32_array (_cs->resp, (a), (n), (mx)); \
+	if (!n) { _ret = CKR_ARGUMENTS_BAD; goto _cleanup; } \
+	if (!a) _cs->overflowed++; \
+	_ret = proto_read_uint32_array (_cs->resp, (a), (n), (mx), &_cs->overflowed); \
 	if (_ret != CKR_OK) goto _cleanup;
 
 #define OUT_RETURN_CODE() { \
